import apt
import apt_pkg
import logging
import os
import pickle
import re
import string
import subprocess
import sys
import urllib

def dumbExtractControl(file, control="control"):
    """ return the control information of a given deb file
        :param file: The deb package file
        :param control: The control information from the deb file.
                        This can be control,{pre,post}{inst,rm}
    """
    from subprocess import Popen, PIPE
    control = Popen(["dpkg-deb","-I",file, control],
                    stdout=PIPE,stderr=open("/dev/null") ).communicate()[0]
    return control

class ArchiveCrawlerCache(apt.Cache):
    def __init__(self, progress=None, rootdir=None):
        rootdir = os.path.abspath(rootdir)
        for d in ["var/cache/apt/archives/partial",
                  "var/lib/apt/lists/partial",
                  "var/lib/dpkg",
                  "etc/apt/"]:
            if not os.path.isdir(os.path.join(rootdir, d)):
                os.makedirs(os.path.join(rootdir, d))
        dpkgstatus = os.path.join(rootdir,"var/lib/dpkg/")+"status"
        if not os.path.exists(dpkgstatus):
            open(dpkgstatus,"w")
        apt.Cache.__init__(self, progress, rootdir=rootdir)

class ArchiveCrawler(object):

    def __init__(self, aptroot, pooldir, actionsdir, datadir, arch=apt_pkg.Config.Find("APT::Architecture"), debfiles_done_database="debfiles-done.p"):
        self.arch = arch
        apt_pkg.Config.Set("APT::Architecture",arch)
        self.aptroot = aptroot
        self.pooldir = pooldir
        self.actionsdir = actionsdir
        self.cache = ArchiveCrawlerCache(rootdir=self.aptroot)
        self.datadir = datadir
        self.debfiles_done_database = debfiles_done_database
        self._loadDebFilesDone()
        self.callbacks = set()

    def registerCallback(self, c):
        if not callable(c):
            raise Exception("need a callable")
        self.callbacks.add(c)
        
    def _loadDebFilesDone(self):
        # load the "done" file
        p = os.path.join(self.datadir, self.debfiles_done_database)
        if os.path.exists(p):
            self.debfiles_done = pickle.load(open(p))
        else:
            self.debfiles_done = set()

    def _saveDebFilesDone(self):
        f = open(os.path.join(self.datadir, self.debfiles_done_database),"w")
        pickle.dump(self.debfiles_done, f)

    def updateCache(self):
        self.cache.update()
        self.cache.open(apt.progress.OpProgress())

    def runActionsOnDebFile(self, file, name, ver, arch, component):
        if self.actionsdir is None:
            return True
        res = True
        for f in os.listdir(self.actionsdir):
            f = os.path.join(self.actionsdir, f)
            logging.debug("running: '%s'" % f)
            return_code = subprocess.call([f, file, name, ver, arch, component])
            if return_code != 0:
                logging.error("script '%s' returned exit-code '%s' on '%s'" % (f, return_code, file))
                res = False
        return res

    def runCallbacks(self, file, name, ver, arch, component):
        res = True
        for c in self.callbacks:
            res &= c(self, file, name, ver, arch, component)
        return res

    def markDone(self, debfile):
        logging.debug("adding '%s' to debfiles_done" % debfile)
        self.debfiles_done.add(debfile)

    def inspectDeb(self, debfile):
        #logging.debug("inspectDeb %s" % debfile)
        m = re.match(".*/(.*)_(.*)_(.*).deb", debfile)
        pkgname = m.group(1)
        pkgver = urllib.unquote(m.group(2))
        pkgarch = m.group(3)
        # filter out
        # ... by arch
        if pkgarch != "all" and pkgarch != self.arch:
            logging.debug("skipping, wrong arch: '%s' (expected '%s')" % (pkgarch, self.arch))
            return False
        # ... done already
        if debfile in self.debfiles_done:
            logging.debug("skipping, already in debfiles_done '%s'" % debfile)
            return False
        # ... by name
        if not self.cache.has_key(pkgname):
            logging.debug("skipping, not in cache: '%s'" % pkgname)
            return False
        # ... by version
        candVerWithoutEpoch = candVer = self.cache[pkgname].candidateVersion
        if candVer and ":" in candVer:        # epoch
            candVerWithoutEpoch = candVer.split(":")[1]
        if candVerWithoutEpoch != pkgver:
            logging.debug("skipping, version '%s' not in cache ('%s' expected)" % (pkgver, candVerWithoutEpoch))
            return False
        # ... by component (main, universe, ..)
        # the archive may have *two* exactly same versions of a package
        # once in pool/main and once in pool/universe, e.g.:
        # '/srv/archive.ubuntu.com/ubuntu/pool/main/n/nessus-core/nessusd_2.2.9-2_i386.deb'
        # '/srv/archive.ubuntu.com/ubuntu/pool/universe/n/nessus-core/nessusd_2.2.9-2_i386.deb'
        # valid deb
        section = self.cache[pkgname].section
        if not "/" in section:
            component = "main"
        else:
            component = section[0:section.find("/")]
        if not debfile.startswith("%s/%s" % (self.pooldir, component)):
            logging.debug("skipping, compoent does not match (expected '%s' got '%s' "% (component, debfile))
            return False

        # looks like we have a valid ver
        logging.debug("found valid deb: '%s'" % debfile)

        
        # pass the epoch here too, this information is not encoded in the
        # filename
        if (self.runActionsOnDebFile(debfile, pkgname, candVer, pkgarch, component) and
            self.runCallbacks(debfile, pkgname, candVer, pkgarch, component)):
            self.markDone(debfile)
    
    def crawl(self):
        # walk over the pool
        for (root, dirs, files) in os.walk(self.pooldir):
            for f in files:
                if f.endswith(".deb"):
                    self.inspectDeb(os.path.join(root,f))
        self._saveDebFilesDone()

    def findOrphanedFiles(self):
        """ walk over the debfiles_done set and find files that are no
            longer in the pool. Useful for e.g. removing meta-data
            
            Note that debfiles with obsolete versions are not part
            of the orphan list if they are still in the pool (e.g.
            new version pkg_1.0 gets superseeded in dapper by
            pkg_1.1, is still in the pool for breezy, then it will
            *not* be part of this list)
        """
        orphaned = set()
        for debfile in self.debfiles_done:
            if not os.path.exists(debfile):
                orphaned.add(debfile)
            
        return orphaned

    def findObsoletedPackages(self):
        """
        return a set of packagenames that got obsoleted
        since the last run of the archive-crawler or
        never got processed so far or got removed
        """
        obsoleted = set()
        # first find all package names that we had in debfiles_done
        # but were the version no longer matches
        for pkg in self.cache:
            candVer = pkg.candidateVersion
            section = pkg.section
            if "/" in section:
                section = section.split("/")[0]
            else:
                section = "main"
            # we get a version with epoch from the cache, but
            # its stored without the epoch in the pool. so we
            # strip it here
            if ":" in candVer:
                candVer = candVer.split(":")[1]
            if pkg.name.startswith("lib"):
                poolhash = pkg.name[0:4]
            else:
                poolhash = pkg.name[0]
            debfile = "%s/%s/%s/%s/%s_%s_%s.deb" % (self.pooldir,
                                                    section,
                                                    poolhash,
                                                    pkg.name,
                                                    pkg.name,
                                                    candVer,
                                                    pkg.architecture)
            if not debfile in self.debfiles_done:
                obsoleted.add(pkg.name)
        # then search for packages that used to be available but
        # got removed for some reason
        for debfile in self.debfiles_done:
            (name, ver, arch) = string.split(os.path.basename(debfile), "_")
            if not self.cache.has_key(name):
                obsoleted.add(name)
        return obsoleted

    def findOrphanedPackages(self):
        """ 
        walk over the debfiles_done set and find packages that are no
        longer in the distribution (but used to be).

        This will return a set of debfiles with full path to the 
        pool dir that are no longer in the current distribution.
        It may contain e.g. "pool/hello_1.0_all.deb" if that got
        superseeded with "pool/hello_1.1_all.deb"
        
        """
        orphaned = set()
        for debfile in self.debfiles_done:
            if not os.path.exists(debfile):
                orphaned.add(debfile)
                continue
            (name, ver, arch) = string.split(os.path.basename(debfile), "_")
            if not self.cache.has_key(name):
                orphaned.add(debfile)
                continue
            candVer = self.cache[name].candidateVersion
            # we get a version with epoch from the cache, but
            # its stored without the epoch in the pool. so we
            # strip it here
            if ":" in candVer:
                candVer = candVer.split(":")[1]
            if candVer != ver:
                orphaned.add(debfile)
        return orphaned
