import ArchiveCrawler
import os
import os.path
import tempfile
import logging
import shutil
import string

import data_extractor
import data_extractor.alternatives

class CommandDataExtractor(object):

    SUPPORTED_ARCHES = ("i386","amd64")

    def __init__(self, aptroot, pooldir, datadir):
        # init dirs
        self.aptroot = aptroot
        self.pooldir = pooldir
        self.datadir = datadir
        self.cmd_data_f = os.path.join(datadir, "scan.data")
        self.pkgs_transform = {}
        renamecfg = os.path.join(self.datadir,"cmd_rename.cfg")
        if os.path.exists(renamecfg):
            for line in map(string.strip, open(renamecfg).readlines()):
                if line != "" and not line.startswith("#"):
                    (oldname,newname) = string.split(string.strip(line))
                    logging.debug("renaming: %s -> %s" % (oldname,newname))
                    self.pkgs_transform[oldname] = newname


    def find_executables(self, packages):
        """Iterate over packages and produce a tuple
           (program_name, package_name, package_repo, package_arch)
            for each executable found in relevant locations.
        """
        def fix_item(item):
            if item.KIND == "hardlink" and item.name.startswith("."):
                item.name = item.name[1:]
            if not item.name.startswith("/"):
                item.name = "/" + item.name
        for package in packages:
            for item in package.contents:
                fix_item(item)
                if (item.is_executable() and
                    item.KIND in ("file", "symlink", "hardlink")):
                    for prefix in ("/bin/", "/usr/bin/", "/sbin/", "/usr/sbin", "/usr/games"):
                        if item.name.startswith(prefix): 
                            logging.debug("Found executable '%s' in package '%s'" % (os.path.basename(item.name), package.file_name))
                            yield (os.path.basename(item.name), package.name, package.repo, package.arch)
            if package.postinst:
                for alt_name, alt_target in data_extractor.alternatives.get_alternatives(package.postinst):
                    for prefix in ("/bin/", "/usr/bin/", "/sbin/", "/usr/sbin", "/usr/games"):
                        if alt_name.startswith(prefix):
                            logging.debug("Found alternative '%s' in package '%s'" % (os.path.basename(alt_name), package.file_name))
                            yield (os.path.basename(alt_name), package.name, package.repo, package.arch)


    def inspectDeb(self, crawler, filename, pkgname, ver, pkgarch, component):
        logging.debug("processing: %s" % filename)
        try:
            extractor_pkg = data_extractor.deb_package.load(filename)
        except Exception ,e:
            logging.error("error in deb_package.load(): %s, %s" % (filename, e))
            return False
        executables = set()
        for name, pkg, repo, arch in self.find_executables([extractor_pkg]):
            executables.add(name)
        if len(executables) == 0:
            return True
        self._cleanPkg(pkgarch, repo, pkgname)
        # see if we have a override
        if pkgname in self.pkgs_transform:
            pkgname = self.pkgs_transform[pkgname]
        s = "%s|%s|%s|%s\n" % (pkgarch,component,pkgname,",".join(executables))
        logging.info("adding '%s'" % s)
        self.cmd_data.append(s)
        return True

    def _cleanPkg(self, pkgarch, comp, pkgname):
        " remove pkgarch,pkgname from self.cmd_data "
        # Done: what if 
        #        a) pkg moves from main->universe (or vice versa)
        #        -> gets detected by findOrphanedPackages() too because
        #           the file is no longer in the old path (that contained
        #           universe)
        #        b) pkg moves from $arch to "all" (or vice versa)
        #        -> requires new upload, gets caught by findOrphanedPackages()
        #        c) pkg gets removed from distro (hardy) but not 
        #           from previous distro (gutsy) so that the orphan
        #           cleanup can not kick in
        #        -> gets caught by findOrphanedPackages()
        for line in self.cmd_data:
            if line.startswith("%s|%s|%s|" % (pkgarch, comp, pkgname)):
                logging.debug("_cleanPkg: %s|%s|%s success" % (pkgarch, comp, pkgname))
                self.cmd_data.remove(line)
                             
    def _cleanOrphans(self):
        if not os.path.exists(self.cmd_data_f):
	    logging.debug("no %s file found, no orphans cleanup" % self.cmd_data_f)
            return
        # build orphaned pkgs set
        # FIXME:
        #  deal with orphans, when we have:
        #    'pool/hello_1.0_all.deb'
        #    'pool/hello_1.1_all.deb' 
        #  in debfiles_done. And hello_1.1 is in the distribution,
        #  but hello_1.0 is no longer. Currently we get 
        #  pool/hello_1.0_all.deb as a orphan and because we strip
        #  away the version information, we kill it in _cleanPkg()
        #  from our records. 
        #  -> But we shouldn't because 1.1 is done but still current
        #  (this probably applies to the desktop extractor too)
        orphaned_pkgs = self.crawler.findObsoletedPackages()

        logging.debug("orphans: %s" % orphaned_pkgs)
        # check if pkg is in current dataset
        t = tempfile.NamedTemporaryFile()
        for line in open(self.cmd_data_f):
            (arch,section,pkg,files) = line.split("|")
            if pkg in orphaned_pkgs:
		logging.debug("removing %s as orphan" % pkg)
                continue
            t.write(line)
        t.flush()
        shutil.copy(t.name, self.cmd_data_f)
	logging.debug("finishing orphan cleanup")

    def extract(self):
        # orphan cleanup code disabled
        #logging.info("Cleanup orphans")
        #self.crawler = ArchiveCrawler.ArchiveCrawler(self.aptroot,
        #                                             self.pooldir,
        #                                             None,
        #                                             self.datadir,
        #                                             debfiles_done_database="command.p")
        #self._cleanOrphans()

        logging.info("Starting extraction in %s" % self.pooldir)
        self.cmd_data = []
        #if os.path.exists(self.cmd_data_f):
        #    self.cmd_data = open(self.cmd_data_f).readlines()
        for arch in self.SUPPORTED_ARCHES:
            self.crawler = ArchiveCrawler.ArchiveCrawler(self.aptroot,
                                                         self.pooldir,
                                                         None,
                                                         self.datadir,
                                                         arch,
                                                         debfiles_done_database="command.p")
            self.crawler.updateCache()
            self.crawler.registerCallback(self.inspectDeb)
            self.crawler.crawl()
        open(self.cmd_data_f,"w").write("".join(self.cmd_data))
        logging.info("extract() finished")

            
