# (c) Zygmunt Krynicki 2005, 2006, 2007
# Licensed under GPL, see COPYING for the whole text

class GenericFileInfo(object):
    """The base class for a simple hierarhy of file-information classes.
    This class serves as a data exchange format of various pargs of the
    scanner. Each instance contains information about a single filesystem entry"""

    BASE_STICKY  = 8
    BASE_READ    = 4
    BASE_WRITE   = 2
    BASE_EXEC    = 1

    O_STICKY= BASE_STICKY << 0
    O_READ  = BASE_READ   << 0
    O_WRITE = BASE_WRITE  << 0
    O_EXEC  = BASE_EXEC   << 0
  
    G_STICKY= BASE_STICKY << 3
    G_READ  = BASE_READ   << 3
    G_WRITE = BASE_WRITE  << 3
    G_EXEC  = BASE_EXEC   << 3
  
    U_STICKY= BASE_STICKY << 6
    U_READ  = BASE_READ   << 6
    U_WRITE = BASE_WRITE  << 6
    U_EXEC  = BASE_EXEC   << 6

    STICKY = O_STICKY  | G_STICKY | U_STICKY
    READ   = O_READ    | G_READ   | U_READ
    WRITE  = O_WRITE   | G_WRITE  | U_WRITE
    EXEC   = O_EXEC    | G_EXEC   | U_EXEC

    __slots__ = ["name", "mode", "uid", "gid", "size", "mtime", "target", "major", "minor"]
  
    def __init__(self, name, mode, uid, gid, size, mtime):
        self.name = name
        self.mode = mode
        self.uid = uid
        self.gid = gid
        self.size = size
        self.mtime = mtime
    def __str__(self):
        return "mode:%s, name:%r, kind:%s" % (self.unix_mode(), self.name, self.KIND)

    def is_executable(self):
        return self.mode & self.EXEC

    def is_readable(self):
        return self.mode & self.READ

    def is_writable(self):
        return self.mode & self.WRITE
  
    def unix_mode(self):
        """ Returns a ls -l like mode string. """
        def mode_str(mode):
            out = ""
            if mode & self.BASE_READ:
                out += 'r'
            else:
                out += '-'
            if mode & self.BASE_WRITE:
                out += 'w'
            else:
                out += '-'
            if mode & self.BASE_EXEC:
                out += 'x'
            else:
                out += '-'
            return out
        def kind_str(self):
            if isinstance(self, FileInfo) or isinstance(self, HardLinkInfo):
                return '-'
            elif isinstance(self, DirectoryInfo):
                return 'd'
            elif isinstance(self, SymbolicLinkInfo):
                return 'l'
            else:
                return '?'
        return "%s%s%s%s" % (kind_str(self), mode_str(self.mode >> 6 & 7), mode_str(self.mode >> 3 & 7), mode_str(self.mode & 7))

class FileInfo(GenericFileInfo):
    KIND = "file"
    def __init__(self, name, mode, uid, gid, size, mtime):
        GenericFileInfo.__init__(self, name, mode, uid, gid, size, mtime)

class DirectoryInfo(GenericFileInfo):
    KIND = "directory"
    def __init__(self, name, mode, uid, gid, size, mtime):
        GenericFileInfo.__init__(self, name, mode, uid, gid, size, mtime)

class LinkInfo(GenericFileInfo):
    (SOFT,HARD) = range(2)
    def __init__(self, name, target, mode, uid, gid, size, mtime):
        GenericFileInfo.__init__(self, name, mode, uid, gid, size, mtime)
        self.target = target
    def get_link_type(self):
        raise NotImplemented
    def __src__(self):
        return "mode:%s, name:%r, target:%r, kind:%s" % (self.unix_mode(), self.name, self.target, self.KIND)

class SymbolicLinkInfo(LinkInfo):
    KIND = "symlink"
    def __init__(self, name, target, mode, uid, gid, size, mtime):
        LinkInfo.__init__(self, name, target, mode, uid, gid, size, mtime)
    def get_link_type(self):
        return self.SOFT

class HardLinkInfo(LinkInfo):
    KIND = "hardlink"
    def __init__(self, name, target, mode, uid, gid, size, mtime):
        LinkInfo.__init__(self, name, target, mode, uid, gid, size, mtime)
    def get_link_type(self):
        return self.HARD

class DeviceInfo(GenericFileInfo):
    (CHAR,BLOCK) = range(2)
    def __init__(self, name, mode, uid, gid, size, mtime, major, minor):
        GenericFileInfo.__init__(self, name, mode, uid, gid, size, mtime)
        self.major = major
        self.minor = minor
    def get_device_type(self):
        pass

class CharDeviceInfo(DeviceInfo):
    KIND = "char device"
    def __init__(self, name, mode, uid, gid, size, mtime, major, minor):
        DeviceInfo.__init__(self, name, mode, uid, gid, size, mtime, major, minor)
    def get_device_type(self):
        return self.CHAR

class BlockDeviceInfo(DeviceInfo):
    KIND = "block device"
    def __init__(self, name, mode, uid, gid, size, mtime, major, minor):
        DeviceInfo.__init__(self, name, mode, uid, gid, size, mtime, major, minor)
    def get_device_type(self):
        return self.BLOCK

class FifoInfo(GenericFileInfo):
    KIND = "fifo"
    def __init__(self, name, mode, uid, gid, size, mtime):
        GenericFileInfo.__init__(self, name, mode, uid, gid, size, mtime)
