
import os
import tarfile
import sys
import apt
import apt_pkg
import apt_inst
import os.path
import re
import tempfile
import subprocess
import string
import shutil
import urllib
import logging
import fnmatch
import glob
import pickle

import ArchiveCrawler

class DesktopDataExtractor(object):

    SUPPORTED_ARCHES = ("i386","amd64")

    def __init__(self, aptroot, pooldir, datadir):
        # init dirs
        self.aptroot = aptroot
        self.pooldir = pooldir
        self.datadir = datadir
        self.tmpdir  = tempfile.mkdtemp()
        self.menu_data = os.path.join(self.datadir, "menu-data")
        self.menu_data_codecs = os.path.join(self.datadir, "menu-data-codecs")
        if not os.path.exists(self.menu_data_codecs):
            raise Exception, "no menu-data-codecs/ dir in %s" % datadir
            
        self.codecs_foradditional = { }
        # packages we have already seen
        self.pkgs_seen = set()
        # pkgs that shouldn't be written (e.g. gnome-about)
        self.pkgs_blacklisted = set()
        # desktop files that are not wanted
        self.desktop_blacklist = set()
        # a mapping to transform certain detections to a new name
        # (e.g.abiword-plugins -> abiword-gnome)
        self.pkgs_transform = {}
        # a mapping from desktop file to a list of free annotations that will
        # be added to the desktop file (useful for e.g. X-AppInstall-Replaces
        self.desktop_annotate = {}
        # pkg to desktop file mapping (this assumes that for a packagename
        # is uniq across architecutures)
        self.pkgs_to_desktop = {}
        # a dictionary with "$arch" -> set() of packages mapping
        # this makes it easy to analyse what packages are only
        # available in certain arches
        self.pkgs_per_arch = {}
        self.pkgs_per_arch["all"] = set()
        # now read the config
        self._readConfig()

        # deb_to_files
        self.deb_to_files_f = os.path.join(self.datadir, "deb_to_files.p")
        if os.path.exists(self.deb_to_files_f):
            self.deb_to_files = pickle.load(open(self.deb_to_files_f))
        else:
            self.deb_to_files = {}

    def __del__(self):
        # be nice, clean-up
        if self.tmpdir != None and self.tmpdir != "":
            shutil.rmtree(self.tmpdir)

    def _readConfig(self):
        # read configfiles
        blacklist = os.path.join(self.datadir,"blacklist.cfg")
        blacklist_desktop = os.path.join(self.datadir,"blacklist_desktop.cfg")
        renamecfg = os.path.join(self.datadir,"rename.cfg")
        annotatecfg = os.path.join(self.datadir,"annotate.cfg")
        if os.path.exists(blacklist):
            logging.info("using blacklist: '%s'" % blacklist)
            for line in open(blacklist).readlines():
                line = line.strip()
                if line != "" and not line.startswith("#"):
                    logging.debug("blacklisting: '%s'" % line.strip())
                    self.pkgs_blacklisted.add(line.strip())

        if os.path.exists(blacklist_desktop):
            logging.info("using blacklist desktop: '%s'" % blacklist_desktop)
            for line in open(blacklist_desktop).readlines():
                line = line.strip()
                if line != "" and not line.startswith("#"):
                    logging.debug("blacklisting (desktop file): '%s'" % line.strip())
                    self.desktop_blacklist.add(line.strip())

        if os.path.exists(renamecfg):
            logging.info("using rename: '%s'" % renamecfg)
            for line in open(renamecfg).readlines():
                line = line.strip()
                if line != "" and not line.startswith("#"):
                    (oldname,newname) = string.split(string.strip(line))
                    logging.debug("renaming: %s -> %s" % (oldname,newname))
                    self.pkgs_transform[oldname] = newname

        if os.path.exists(annotatecfg):
            logging.info("using annotates: '%s'" % annotatecfg)
            for line in open(annotatecfg):
                line = line.strip()
                if line != "" and not line.startswith("#"):
                    (desktopfile,annotations_str) = string.split(line,":")
                    annotations = annotations_str.split(",")
                    logging.debug("annotations: '%s': %s" % (desktopfile,annotations))
                    self.desktop_annotate[desktopfile] = annotations



    def extract(self):
        logging.info("Cleanup orphans")
        self.crawler = ArchiveCrawler.ArchiveCrawler(self.aptroot,
                                                     self.pooldir,
                                                     None,
                                                     self.datadir,
                                                     debfiles_done_database="desktop.p")
        self._cleanOrphans()

        logging.info("Starting extraction in %s" % self.pooldir)
        for arch in self.SUPPORTED_ARCHES:
            self.pkgs_per_arch[arch] = set()
            self.crawler = ArchiveCrawler.ArchiveCrawler(self.aptroot,
                                                         self.pooldir,
                                                         None,
                                                         self.datadir,
                                                         arch,
                                                         debfiles_done_database="desktop.p")
            self.crawler.updateCache()
            self.crawler.registerCallback(self.inspectDeb)
            self.crawler.crawl()
        self._calcArchSpecific()
        self._addCodecInformation()
        pickle.dump(self.deb_to_files,open(self.deb_to_files_f,"w"))
        logging.info("extract() finished")
        
    def _calcArchSpecific(self):
        # now add the architecture information
        arch_specific = set()
        for pkg in self.pkgs_seen:
            arches = []
            if not pkg in self.pkgs_per_arch["all"]:
                for arch in self.SUPPORTED_ARCHES:
                    if not pkg in self.pkgs_per_arch[arch]:
                        arch_specific.add(pkg)
                    else:
                        arches.append(arch)

            if pkg in arch_specific:
                str="X-AppInstall-Architectures=%s\n" % ",".join(arches)
                if not self.pkgs_to_desktop.has_key(pkg):
                    logging.debug("Skipping arch-specific pkg '%s' with no desktop file" % pkg)
                    continue
                for f in self.pkgs_to_desktop[pkg]:
                    dfile = os.path.join(self.menu_data,f)
                    if os.path.exists(dfile):
                        logging.debug("adding arch-specifc information to %s (%s)" % (dfile,arches))
                        open(dfile,"a").write(str)
                    else:
                        logging.error("can't find desktopfile for: %s (%s)" % (pkg,dfile))

    def _addCodecInformation(self):
        # edit the codec lists into the entries in menu-data-additional
        logging.debug("adding codecs: '%s'", self.codecs_foradditional)
        seen = set()
        for filename in glob.glob(self.menu_data_codecs+'/*.desktop'):
            lines = []
            file = open(filename)
            pkgname = None
            for line in file:
                line = string.strip(line)
                if re.match("^X-AppInstall-Codecs\s*=", line): continue
                lines.append(line)
                m = re.match("^X-AppInstall-Package\s*=\s*(\S+)\s*$", line)
                if m: pkgname = m.group(1)
            file.close()
            if not pkgname:
                logging.warning("file '%s' doesn't specify an X-AppInstall-Package" %
                                filename)
                continue
            try:
                codecs = self.codecs_foradditional[pkgname]
            except KeyError:
                continue
            logging.debug("adding codec info to '%s' " % pkgname)
            lines.append("X-AppInstall-Codecs=%s" % ';'.join(codecs))
            seen.add(pkgname)
            new_file = open(filename+'.tmp', 'w');
            for line in lines: print >>new_file, line
            new_file.close()
            os.rename(filename+'.tmp', filename)

        missing = set(self.codecs_foradditional.keys())-seen
        if len(missing) > 0:
            logging.warning("Packages with codecs but no data in menu-data-additonal: %s" % missing)


    def _getMemberFromAr(self, arFile, member):
        """ helper to extract the files """
        if not os.path.exists(arFile):
            raise IOError("file '%s' does not exit" % arFile)
        arFile = os.path.abspath(arFile)
        dataPath = self.tmpdir
        olddir = os.getcwd()
        os.chdir(dataPath)
        command = ["ar","x",arFile, member]
        subprocess.call(command)
        memberpath = dataPath+"/"+member
        os.chdir(olddir)
        if os.path.exists(memberpath):
            return memberpath
        else:
            return None

    def extract_icon(self, tarfile, iconName, newIconName):
        logging.info("extract_icon: %s %s %s" % (tarfile.name, iconName, newIconName))
        extractName = iconName
        if iconName.startswith('/'):
            extractName = ".%s" % iconName
        try:
            iconFile = tarfile.extractfile(extractName)
            outicon = open(newIconName, "w")
            outicon.write(iconFile.read())
            outicon.close()
            iconFile.close()
            logging.debug("wrote iconfile '%s' (from '%s') " % (os.path.basename(outicon.name),iconName))
            return True
        except Exception,e:
            # we may sometimes get very confusing errors from tarfile here
            # (like 'filename None not found' from xmms)
            #  - this usually means something is strange in the
            # tarball eg. xmms.xpm is a symlink to the (non-existant)
            # xmms_mini.xpm
            logging.error("ERROR: Icon '%s' for could not be obtained: %s " % (iconName,e))
        return False

    def search_icon(self, tarfile, iconName, outputdir):
        if iconName == None:
            logging.warning("search_icon() called with no icon name")
            return (False, None)

        # a iconName can be a single name or a full path
        # if it is a single name, look into a icon-theme path (usr/share/icons/hicolor) and then into usr/share/pixmaps
        # if it is a full path just look for this

        # this is the "full-path" case
        # FIXME: there are (some) icons that are not full pathes like "/usr/.../"
        #        but "zapping/zapping.png"
        if "/" in iconName:
            newIconName = iconName.replace("/", "_")
            res = self.extract_icon(tarfile, iconName, os.path.join(outputdir,"icons",newIconName))
            return (res, newIconName)

        # this is the "get-it-from-a-icontheme" case, look into icon-theme hicolor and usr/share/pixmaps

        # search path (ordered by importance)
        search_dirs = [
            "usr/share/icons/hicolor/64x64",
            "usr/share/icons/hicolor/48x48",
            "usr/share/icons/hicolor/128x128",
            "usr/share/pixmaps",
            "usr/share/icons/hicolor/32x32",
            "usr/share/icons/hicolor/22x22",
            "usr/share/icons/hicolor/16x16",
            "usr/share/icons"
            ]
        # extensions (ordered by importance)
        pixmaps_ext = ["", ".png",".xpm",".svg"]

        for d in search_dirs:
            for name in tarfile.getnames():
                if d in name:
                    for ext in pixmaps_ext:
                        if name.endswith(iconName+ext):
                            res = self.extract_icon(tarfile, name, os.path.join(outputdir,"icons", os.path.basename(name)))
                            return (res, None)
        logging.warning("no icon: '%s' could be found" % iconName)
        return (False, None)
    
    def tarfile_extract_orlog(self, dataFile, path):
        try:
            file = dataFile.extractfile(path)
        except KeyError, e:
            logging.error("Failed to extract '%s' from '%s' (%s)" %
                          (path, dataFile.name, e))
            return None
        return file

    def writeMetaData(self, outfile, pkgname, section, codecs):
        outfile.write("X-AppInstall-Package=%s\n" % pkgname)
        outfile.write("X-AppInstall-Section=%s\n" % section)
        if codecs:
            outfile.write("X-AppInstall-Codecs=%s\n" % ';'.join(codecs))
        outfile.write("\n")
        return True

    def getFiles(self, dataPath, pkgname, section, debfile):
        """ 
        actually extract the files from a data.tar.[gz,bz2,lzma}
        of a package
        """
        outputdir = self.menu_data
        try:
            os.makedirs(os.path.join(outputdir, "icons"))
        except OSError:
            pass

        try:
            dataFile = tarfile.open(dataPath)
        except:
            logging.error("Couldn't open tarball. Package: '%s'" % dataPath)
            return

        desktopPaths = []

        # get the available desktop files
        try:
            for filename in dataFile.getnames():
                if (filename.endswith(".desktop") and 
                   not os.path.basename(filename) in self.desktop_blacklist and 
                    "usr/share/applications" in filename):
                    desktopPaths.append(filename)
        except:
            logging.error("Choked on tarball. Package: '%s'" % package)
            return # too many bad debs. take this line out if you fix that

        # and look for any codec support list
        codecs = set()
        for filename in dataFile.getnames():
            if not (fnmatch.fnmatch(filename, 'usr/share/gstreamer-*/plugin-info/*.supported') or fnmatch.fnmatch(filename, './usr/share/gstreamer-*/plugin-info/*.supported')): continue
            codecsfile = self.tarfile_extract_orlog(dataFile,filename);
            if not codecsfile:
                continue
            logging.debug("reading codec list file %s" % filename)
            for line in codecsfile.readlines():
                line = string.strip(line)
                line = string.rstrip(line, ";") # remove ";" at the end
                if ';' in line:
                    logging.warning("Codec string '%s' from '%s' in '%s' contains ';'" %
                                    (line, filename, dataFile.name))
                    continue
                codecs.add(line)

        # look at the desktop files
        for path in desktopPaths:
            desktopfile = self.tarfile_extract_orlog(dataFile, path)
            if not desktopfile: continue
            # new location for the desktopfile, add prefix if the 
            # if the desktop file is in a subdir, encode that into
            # the name
            prefix = ""
            (dir, name) = os.path.split(path)
            dir = dir.rstrip("/")
            if not dir.endswith("applications"):
                # FIXME this will not work well if there are multiple 
                # subdirs under applications/
                prefix = dir.split("/")[-1] + "_"
            desktop = os.path.join(outputdir, prefix+os.path.basename(path))
            # extract the icon
            iconName = None
            newIconName = None
            for line in desktopfile.readlines():
                line = string.strip(line)
                if line.startswith("Icon="):
                    iconName = line[line.index("=")+1:]
                    logging.debug("Package '%s' needs icon '%s'" % (pkgname, iconName))
            (res, newIconName) = self.search_icon(dataFile, iconName, outputdir)

            # now check for supicious pkgnames (FIXME: make this not hardcoded)
            if "-common" in pkgname or "-data" in pkgname:
                logging.warning("'%s' looks wrong, trying to correct" % pkgname)
                pkg = self.crawler.cache[pkgname]
                parentpkg = pkgname[0:pkgname.rindex("-")]
                if self.crawler.cache.has_key(parentpkg):
                    logging.warning("Corrected to '%s'" % parentpkg)
                    pkgname = parentpkg

            # now write out the file
            arch = apt_pkg.Config.Find("APT::Architecture")
            logging.debug("Writing desktop file '%s' for arch '%s'" % (desktopfile.name, arch))
            outfile = open(desktop, "w")
            desktopfile.seek(0)
            written=False
            desktop_section = ""
            for line in desktopfile.readlines():
                if newIconName != None and line.startswith("Icon="):
                    # Using the file extension is not themable
                    if os.path.splitext(newIconName)[1] != "":
                        line = "Icon=%s\n" % os.path.splitext(newIconName)[0]
                    else:
                        line = "Icon=%s\n" % newIconName
                if not line.endswith("\n"):
                    line += "\n"
                # if a new section starts, write the metadata now
                if desktop_section == "[Desktop Entry]" and not written:
                    self.writeMetaData(outfile, pkgname, section, codecs)
                    written=True
                outfile.write(line)
                if line.startswith("["):
                    desktop_section = line.strip()
            if not written:
                self.writeMetaData(outfile, pkgname, section, codecs)

            desktopfile_name = os.path.basename(desktopfile.name)
            if self.desktop_annotate.has_key(desktopfile_name):
                logging.info("adding annontate '%s' for '%s'" % (self.desktop_annotate[desktopfile_name], desktopfile_name))
                for a in self.desktop_annotate[desktopfile_name]:
                    outfile.write("%s\n" % a.strip())
            outfile.close()
            # record pkgname->desktopfile
            if not self.pkgs_to_desktop.has_key(pkgname):
                self.pkgs_to_desktop[pkgname] = set()
            self.pkgs_to_desktop[pkgname].add(os.path.basename(outfile.name))
            # and debfile->desktopfile
            if not self.deb_to_files.has_key(debfile):
                self.deb_to_files[debfile] = set()
            self.deb_to_files[debfile].add(os.path.basename(outfile.name))
            
            # close the desktop file
            desktopfile.close()
            logging.info("added %s" % desktopfile.name)

        if codecs:
            try: self.codecs_foradditional[pkgname] &= codecs
            except KeyError: self.codecs_foradditional[pkgname] = codecs

    def processDeb(self, debPath, pkgname, section):
        """ extract the desktop file and the icons from a deb """
        outputdir=self.menu_data
        logging.debug("processing: %s" % debPath)
        datafile = self._getMemberFromAr(debPath, "data.tar.gz")
        if datafile == None:
            datafile = self._getMemberFromAr(debPath, "data.tar.bz2")
            if datafile == None:
                datafile = self._getMemberFromAr(debPath, "data.tar.lzma")
                if datafile == None:
                    logging.error("error getting data.tar from %s" % debPath)
                    return
                else:
                    # extract it here, python tarfile does not support lzma
                    subprocess.call(["lzma","-d",datafile])
                    datafile = os.path.splitext(datafile)[0]
        self.getFiles(datafile, pkgname, section, debPath)
        os.remove(datafile)

    def inspectDeb(self, crawler, filename, pkgname, ver, pkgarch, component):
        """ check if the deb is interessting for us (not blacklisted) """
        logging.debug("inspectDeb '%s'"% filename)
        # add it to the arch-table now, even if it might be known already
        self.pkgs_per_arch[pkgarch].add(pkgname)

        if pkgname in self.pkgs_blacklisted:
            logging.warning("skipping blacklisted pkg: '%s'" % pkgname)
            return True
        if self.pkgs_transform.has_key(pkgname):
            logging.warning("transforming '%s' to '%s'" % (pkgname, self.pkgs_transform[pkgname]))
            pkgname = self.pkgs_transform[pkgname]

        # found somethat worth looking at
        self.processDeb(filename, pkgname, component)
        self.pkgs_seen.add(pkgname)
        return True

    def _cleanOrphans(self):
        " this must be run *before* a extraction, never after "
        orphans = self.crawler.findOrphanedFiles()
        for orphan in orphans:
            if self.deb_to_files.has_key(orphan):
                logging.debug("found orphaned '%s'->'%s'" % (orphan,self.deb_to_files[orphan]))
                for f in self.deb_to_files[orphan]:
                    p = os.path.join(self.menu_data,f)
                    if os.path.exists(p):
                       logging.debug("removing '%s'" % os.path.join(self.menu_data,f))
                       os.unlink(os.path.join(self.menu_data,f))
                    else:
                        logging.debug("can not find '%s' in %s" % (f, self.menu_data))
